/*
Copyright (c) 2014 Carnegie Mellon University.
All Rights Reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following acknowledgments and disclaimers.
2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following acknowledgments and disclaimers in the documentation and/or other materials provided with the distribution.
3. Products derived from this software may not include Carnegie Mellon University, "SEI and/or Software Engineering Institute" in the name of such derived product, nor shall Carnegie Mellon University, "SEI and/or Software Engineering Institute" be used to endorse or promote products derived from this software without prior written permission. For written permission, please contact permission@sei.cmu.edu.
ACKNOWLEDMENTS AND DISCLAIMERS:
Copyright 2014 Carnegie Mellon University

This material is based upon work funded and supported by the Department of Defense under Contract No. FA8721-05-C-0003 with Carnegie Mellon University for the operation of the Software Engineering Institute, a federally funded research and development center.

Any opinions, findings and conclusions or recommendations expressed in this material are those of the author(s) and do not necessarily reflect the views of the United States Department of Defense.

NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING INSTITUTE MATERIAL IS FURNISHED ON AN AS-IS BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.

This material has been approved for public release and unlimited distribution.

Carnegie Mellon is registered in the U.S. Patent and Trademark Office by Carnegie Mellon University.

DM-0000890

 */

package edu.cmu.sei.mCOP;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.TreeSet;
import java.util.Vector;

import edu.cmu.sei.ZeroSlackBinPacker.PTask;
import edu.cmu.sei.ZeroSlackRM.CriticalZoneTaskIterable;
import edu.cmu.sei.ZeroSlackRM.CriticalityComparator;
import edu.cmu.sei.ZeroSlackRM.DualExecutionZone;
import edu.cmu.sei.ZeroSlackRM.PriorityComparator;
import edu.cmu.sei.ZeroSlackRM.RMZoneTaskIterable;
import edu.cmu.sei.ZeroSlackRM.SlackRegion;
import edu.cmu.sei.ZeroSlackRM.Task;
import edu.cmu.sei.ZeroSlackRM.ZSRMScheduler;

public class TransitionZSRMScheduler extends ZSRMScheduler {
	
	long maxCriticality;

	TreeSet<MTask> mTasksByCriticality = new TreeSet<MTask>(new CriticalityComparator());
	TreeSet<MTask> mTasksByPriority = new TreeSet<MTask>(new PriorityComparator());
	
	HashMap<Mode, TreeSet<Task>> modalTasksetByPriority = new HashMap<Mode,TreeSet<Task>>();
	
	public void addTask(MTask t){
		mTasksByCriticality.add(t);
		mTasksByPriority.add(t);
	}
	
	public void addAllTasks(Collection<MTask> tasks){
		for (MTask t:tasks){
			addTask(t);
		}
	}
	/*
	 * Using the modal facade may not work property because when we calculate the response time of a task i
	 * we check in the iterators if the current task is the task i itself
	 */
	public TreeSet<Task> getModeTasksByPriority(Mode m){
		TreeSet<Task> taskset = modalTasksetByPriority.get(m);
		if (taskset ==null){
			taskset = new TreeSet<Task>(new PriorityComparator());
			modalTasksetByPriority.put(m,taskset);
			for (MTask tsk:mTasksByPriority){
				Task mFacade =tsk.modalFacade.get(m); 
				if ( mFacade != null){
					taskset.add(mFacade);
				}
			}
		}
		
		return taskset;
	}

	public TreeSet<MTask> getPersistentTasksByCriticality(Mode s, Mode t){
		TreeSet<MTask> persistentTasksByCriticality=new TreeSet<MTask>(new CriticalityComparator());

		for (MTask tsk:mTasksByPriority){
			if (tsk.modalFacade.get(s) != null && tsk.modalFacade.get(t)!=null){
				persistentTasksByCriticality.add(tsk);
			}			
		}
		return persistentTasksByCriticality;
	}
	
	public long getTrailingSlackFromInstant(long instant, Vector<SlackRegion> slackVector){
		long slack=0;
		
		// skip slack before instant
		
		Iterator<SlackRegion> regionIterator = slackVector.iterator();
		SlackRegion r =null;
		if (regionIterator.hasNext()){
			r = regionIterator.next();
			while ((r.getTimeInstant()+r.getSlack() < instant) && regionIterator.hasNext()){
				// skip
				r = regionIterator.next();
			}
			if (r.getTimeInstant()+r.getSlack() > instant){
				// some slack
				if (r.getTimeInstant() <= instant ){
					// potentailly partial slack
					slack += (r.getSlack() - (instant - r.getTimeInstant()));
				} else {
					// r.getTimeInstant() > instant
					// all slack
					slack += r.getSlack(); 
				}
			}
			// get rest of slack
			while (regionIterator.hasNext()){
				r = regionIterator.next();
				slack += r.getSlack();
			}
		}
		
		return slack;
	}
	
	public Vector<SlackRegion> buildHybridSlackVector(Vector<SlackRegion> leadingRegion, Vector<SlackRegion> trailingRegion, long switchingPoint){
		Vector<SlackRegion> hybridVector = new Vector<SlackRegion>();
		 
		long endPrevSlack =0;
		for (SlackRegion r:leadingRegion){
			if (r.getTimeInstant()+r.getSlack()<= switchingPoint){
				// add all to new vector
				hybridVector.add(r);
				endPrevSlack = r.getTimeInstant()+r.getSlack();
			} else if (r.getTimeInstant() < switchingPoint){
				SlackRegion r1 = new SlackRegion(switchingPoint,(r.getSlack() - (switchingPoint - r.getTimeInstant()) ));
				hybridVector.add(r1);
				endPrevSlack = r1.getTimeInstant()+r1.getSlack();
			} // else if getTimeInstant == switchingPoint => no slack 
			//else 	// r.getTimeInstant() > switchingPoint// skip nothing to do?
		}
		
		for (SlackRegion r:trailingRegion){
			// skip all previous to switchingPoint
			if (r.getTimeInstant()+r.getSlack() > switchingPoint){
				if (r.getTimeInstant() >= switchingPoint){
					// add all the slack
					hybridVector.add(r);
				} else {
					// getTimeInstant() < switchingPoint
					SlackRegion r1 = new SlackRegion(switchingPoint,r.getSlack() - (switchingPoint - r.getTimeInstant()) );
					hybridVector.add(r1);
				}
			} // else -- skip
		}
		return hybridVector;
	}
	
	public long getMaximizingMCRI(Task taski, Mode s, Mode t){
		long maxMrci=0;
		
		SlackRegion infRegion = new SlackRegion(Long.MAX_VALUE,0);
		
		Task t_s = ((MTask)taski).getModalFacade(s);
		Task t_t = ((MTask)taski).getModalFacade(t);
		
		Vector<SlackRegion> srcRMSlackRegion  = calculateSlackVector(t_s, new RMZonePlusHigherCriticalIterable(getModeTasksByPriority(s),t_s));
		Vector<SlackRegion> tgtRMSlackRegion  = calculateSlackVector(t_t, new RMZonePlusHigherCriticalIterable(getModeTasksByPriority(t),t_t));
	
		long currentInstant = 0;
		Iterator<SlackRegion> srcRegionIterator = srcRMSlackRegion.iterator();
		Iterator<SlackRegion> tgtRegionIterator = tgtRMSlackRegion.iterator();
		
		if (srcRegionIterator.hasNext() && tgtRegionIterator.hasNext()){
			SlackRegion srcRegion = srcRegionIterator.next();
			SlackRegion tgtRegion = tgtRegionIterator.next();
			//getAvailableSlackUpToInstant(long instant, Vector<SlackRegion> slackV)
			//getTrailingSlackInstant(Task task, long slack, Vector<SlackRegion> slackV)
			// obtain all increasing point from both modes
			ArrayList<Long> slackStartingPoints = new ArrayList<Long>();
			while (currentInstant < taski.getPeriod()){
				if (srcRegion.getTimeInstant()<tgtRegion.getTimeInstant()){
					// next point
					currentInstant = srcRegion.getTimeInstant();
					slackStartingPoints.add(srcRegion.getTimeInstant());
					if (srcRegionIterator.hasNext()){
						srcRegion = srcRegionIterator.next();
					} else {
						srcRegion = infRegion;
					}
				} else {
					currentInstant = tgtRegion.getTimeInstant();
					slackStartingPoints.add(tgtRegion.getTimeInstant());
					if (tgtRegionIterator.hasNext()){
						tgtRegion = tgtRegionIterator.next();
					} else {
						tgtRegion = infRegion;
					}
				}
			}
			slackStartingPoints.add(taski.getPeriod());
			
			// try all 'slack' points as candidate MCIs
			long minSlack = Long.MAX_VALUE;
			long minPoint=0;
			for (long p:slackStartingPoints){
				long slack = getAvailableSlackUpToInstant(p,srcRMSlackRegion)+
							getTrailingSlackFromInstant(p,tgtRMSlackRegion);
				if (slack < minSlack){
					// new min slack candidate
					minSlack = slack;
					minPoint = p;
				}
			}
			maxMrci = minPoint;
		}
				
		return maxMrci;
	}
	/**
	 * We only need to calculate the transitional zero slack for the
	 * persistent task across modes.
	 * 
	 * @param s : source mode
	 * @param t : destination mode
	 */
	public void TransitionCompressSchedule(Mode s, Mode t){
		TreeSet<MTask> persistentTasksByCriticality = getPersistentTasksByCriticality(s,t);
		
		boolean sameZsInstants=false;
		long [] prevZsInstant = new long[persistentTasksByCriticality.size()];
		long [] currZsInstant = new long[persistentTasksByCriticality.size()];

		for (int i=0;i<prevZsInstant.length;i++){
			prevZsInstant[i]=0;
			currZsInstant[i]=0;
		}
		// initialize execution zones
		for (MTask task:persistentTasksByCriticality){
			DualExecutionZone deZone = new DualExecutionZone();
			// use the target (t) mode facade as the 'tested' task
			executionZones.put(task.getModalFacade(t),deZone);
		}

		do {
			for (int i=0;i<prevZsInstant.length;i++){
				prevZsInstant[i] = currZsInstant[i];
			}			

			int index=0;
			for (MTask task:persistentTasksByCriticality){
				// need to 'test' if the mode facade works
				Task tModeFacade = task.getModalFacade(t);
				Task sModeFacade = task.getModalFacade(s);
				TreeSet<Task> ttmp = getModeTasksByPriority(t);
				TreeSet<Task> stmp = getModeTasksByPriority(s);

				// we should use this taskset for the t mode instead. 
				// Need to calculate equivalent taskset for mode s.
				TreeSet<Task> ttasks = new TreeSet<Task>(new PriorityComparator());
				ttasks.addAll(ttmp);
				// add all old tasks (one job per task)
				for (Task tsk:stmp){
					if (!tsk.equals(sModeFacade)){
						PTask pt = (PTask) tsk;
						// double check if this is an 'old' persistent task that must be
						// activated only after its T_j has elapsed (not reflected in the
						// clone.
						Task clone = new PTask(pt.name,tModeFacade.getPeriod(),pt.getOverloadedC(),pt.getNormalC(),pt.criticality,pt.RmPriority);
						ttasks.add(clone);
					}
				}
				// build hybrid transitional slack vector
				long switchingPoint = getMaximizingMCRI(task , s, t);
				
				if (switchingPoint != 0){
					for (MTask t1:persistentTasksByCriticality){
						PTask pts1 = t1.getModalFacade(s);
						PTask ptt1 = t1.getModalFacade(t);
						System.out.println("t(s)[Cr("+pts1.criticality+")T("+pts1.getPeriod()+")C("+pts1.getNormalC()+")Co("+pts1.getOverloadedC()+")] | t(t)[Cr("+ptt1.criticality+")T("+ptt1.getPeriod()+")C("+ptt1.getNormalC()+")Co("+ptt1.getOverloadedC()+")]");
					}
				}
				
				System.out.println("Task("+task.name+") switchingPoint: "+switchingPoint);
				Vector<SlackRegion> srcRMSlackRegion  = calculateSlackVector(sModeFacade, new RMZonePlusHigherCriticalIterable(getModeTasksByPriority(s),sModeFacade));
				Vector<SlackRegion> tgtRMSlackRegion  = calculateSlackVector(tModeFacade, new RMZonePlusHigherCriticalIterable(getModeTasksByPriority(t),tModeFacade));
				Vector<SlackRegion> hybridSlackVector = buildHybridSlackVector(srcRMSlackRegion, tgtRMSlackRegion, switchingPoint);
				
				executionZones.get(tModeFacade).cSlackVector = calculateSlackVector(tModeFacade, new CriticalZoneTaskIterable(ttasks,tModeFacade));//getModeTasksByPriority(t),tModeFacade));
				executionZones.get(tModeFacade).rmSlackVector = hybridSlackVector;//calculateSlackVector(sModeFacade, new RMZoneTaskIterable(getModeTasksByPriority(s),sModeFacade));
				//executionZones.get(sModeFacade).rmSlackVector = hybridSlackVector;//calculateSlackVector(sModeFacade, new RMZoneTaskIterable(getModeTasksByPriority(s),sModeFacade));
				currZsInstant[index] = getZeroSlackInstant(tModeFacade, executionZones.get(tModeFacade).rmSlackVector, executionZones.get(tModeFacade).cSlackVector);
				task.ZeroSlackInstant =currZsInstant[index]; 
				index++;
			}
			
			// calculate finishing condition
			sameZsInstants=true;
			for (int i=0;i<prevZsInstant.length;i++){
				sameZsInstants = sameZsInstants && (prevZsInstant[i] == currZsInstant[i]);
			}
		} while(!sameZsInstants);
		
		// This may not be important in a mode transition. It would destroy the t non-transition 
		// Z.
		int i=0;
		for (MTask task:persistentTasksByCriticality){
			task.transZeroSlackInstant = currZsInstant[i];
			//task.getModalFacade(t).ZeroSlackInstant = currZsInstant[i];
			i++;
		}		

	}

	public void LingueringTransitionCompressSchedule(Mode s, Mode t){
		TreeSet<MTask> persistentTasksByCriticality = getPersistentTasksByCriticality(s,t);
		
		boolean sameZsInstants=false;
		long [] prevZsInstant = new long[persistentTasksByCriticality.size()];
		long [] currZsInstant = new long[persistentTasksByCriticality.size()];

		for (int i=0;i<prevZsInstant.length;i++){
			prevZsInstant[i]=0;
			currZsInstant[i]=0;
		}
		// initialize execution zones
		for (MTask task:persistentTasksByCriticality){
			DualExecutionZone deZone = new DualExecutionZone();
			// use the target (t) mode facade as the 'tested' task
			executionZones.put(task.getModalFacade(t),deZone);
		}

		do {
			for (int i=0;i<prevZsInstant.length;i++){
				prevZsInstant[i] = currZsInstant[i];
			}			

			int index=0;
			for (MTask task:persistentTasksByCriticality){
				// need to 'test' if the mode facade works
				Task tModeFacade = task.getModalFacade(t);
				Task sModeFacade = task.getModalFacade(s);
				TreeSet<Task> ttmp = getModeTasksByPriority(t);
				TreeSet<Task> stmp = getModeTasksByPriority(s);
				// we should use this taskset for the t mode instead. 
				// Need to calculate equivalent taskset for mode s.

				// Now both in nominal and critical state we use the Gamma_t taskset but
				// we add to the nominal state the old tasks that linger during the transitional
				// interval (just one job)
				TreeSet<Task> ttasks = new TreeSet<Task>(new PriorityComparator());
				ttasks.addAll(ttmp);
				TreeSet<Task> stasks = new TreeSet<Task>(new PriorityComparator());
				stasks.addAll(ttmp);
				for (Task tsk:stmp){
					if (!tsk.equals(sModeFacade)){
						PTask pt = (PTask) tsk;
						// double check if this is an 'old' persistent task that must be
						// activated only after its T_j has elapsed (not reflected in the
						// clone.
						Task clone = new PTask(pt.name,tModeFacade.getPeriod(),pt.getOverloadedC(),pt.getNormalC(),pt.criticality,pt.RmPriority);
						stasks.add(clone);
					}
				}

				// We should only evaluate the task in the new mode s
				executionZones.get(tModeFacade).cSlackVector = calculateSlackVector(tModeFacade, new CriticalZoneTaskIterable(getModeTasksByPriority(t),tModeFacade));
				executionZones.get(tModeFacade).rmSlackVector = calculateSlackVector(tModeFacade, new RMZoneTaskIterable(stasks,tModeFacade));//getModeTasksByPriority(s),tModeFacade));
				currZsInstant[index] = getZeroSlackInstant(tModeFacade, executionZones.get(tModeFacade).rmSlackVector, executionZones.get(tModeFacade).cSlackVector);
				task.ZeroSlackInstant =currZsInstant[index]; 
				index++;
			}
			
			// calculate finishing condition
			sameZsInstants=true;
			for (int i=0;i<prevZsInstant.length;i++){
				sameZsInstants = sameZsInstants && (prevZsInstant[i] == currZsInstant[i]);
			}
		} while(!sameZsInstants);
		
		// This may not be important in a mode transition. It would destroy the t non-transition 
		// Z.
		int i=0;
		for (MTask task:persistentTasksByCriticality){
			task.lingeringTransZeroSlackInstant = currZsInstant[i];
			//task.getModalFacade(t).ZeroSlackInstant = currZsInstant[i];
			i++;
		}		

	}

	public double getTransitionalDuctility(Mode s, Mode t, long numCriticalities, long []numTasksPerCriticality, long numTransitions){
		double duct=0.0;
		TreeSet<MTask> persistentTasksByCriticality = getPersistentTasksByCriticality(s,t);
		for (MTask tsk:persistentTasksByCriticality){
			PTask ttask = tsk.getModalFacade(t);
			double tasksPerCrit;
			if (numTasksPerCriticality[(int)ttask.criticality] ==0)
				tasksPerCrit = 1.0;
			else 
				tasksPerCrit = numTasksPerCriticality[(int)ttask.criticality];
			double expDensity =((double)(ttask.getPeriod() - tsk.transZeroSlackInstant) / (double) ttask.getPeriod());
			duct += (expDensity / (Math.pow(2, (maxCriticality*(ttask.criticality-1))) +1.0)/tasksPerCrit);
		}
		return (duct/((double)numCriticalities*numTransitions));
	}
	
	// This is the loss of transitional ductility when using zsrm instead of mzsrm
	public double getProportionalTransitionalDuctilityLoss(Mode s, Mode t, long numCriticalities, long []numTasksPerCriticality, long numTransitions, double[] ductFactorPerCriticality){
		double duct=0.0;
		TreeSet<MTask> persistentTasksByCriticality = getPersistentTasksByCriticality(s,t);
		for (MTask tsk:persistentTasksByCriticality){
			PTask ttask = tsk.getModalFacade(t);
			PTask oldTask = tsk.getModalFacade(s);
			System.out.println("zs: "+oldTask.ZeroSlackInstant+" tzs: "+tsk.transZeroSlackInstant);
			if(oldTask.ZeroSlackInstant > tsk.transZeroSlackInstant){
				duct = (1*ductFactorPerCriticality[(int) oldTask.criticality]);
			}
		}
		return (duct / (double) persistentTasksByCriticality.size());
	}

	public double getProportionOfCriticalityViolations(Mode s, Mode t, long numCriticalities, long []numTasksPerCriticality, long numTransitions, double[] ductFactorPerCriticality){
		long critviol=0;
		TreeSet<MTask> persistentTasksByCriticality = getPersistentTasksByCriticality(s,t);
		for (MTask tsk:persistentTasksByCriticality){
			PTask ttask = tsk.getModalFacade(t);
			if(ttask.ZeroSlackInstant > tsk.transZeroSlackInstant){
				critviol++;
			}
		}
		return ((double)critviol/(double)persistentTasksByCriticality.size());		
	}
	
	
	public static void testTransitionalCompress(){
		final Mode m1 = new Mode();
		final Mode m2 = new Mode();
		
		Vector<Mode> modes = new Vector<Mode>();
		modes.add(m1);
		modes.add(m2);
		//MTask(String name, long T, long Ca, long Cs, HashMap<Mode,Long> criticalities,long RMPriority, numModes)	
		MTask t1 = new MTask("t1",40, 1,1, new HashMap<Mode,Long> () {{put(m1,1l);put(m2,0l);}},3,2);
		MTask t2 = new MTask("t2",10, 5,1, new HashMap<Mode,Long> () {{put(m1,1l);}},1,2);
		MTask t3 = new MTask("t3",30, 3,1, new HashMap<Mode,Long> () {{put(m2,0l);}},2,2);
		
		
		
		TransitionZSRMScheduler tscheduler = new TransitionZSRMScheduler();
		tscheduler.addTask(t1);
		tscheduler.addTask(t2);
		tscheduler.addTask(t3);
		
		tscheduler.TransitionCompressSchedule(m1,m2);

		System.out.println("t1.transz: "+t1.transZeroSlackInstant);
		System.out.println("t2.transz: "+t2.transZeroSlackInstant);
		System.out.println("t3.transz: "+t3.transZeroSlackInstant);
	}
	
	public static void testHybridSlackVector(){
		final Mode m1 = new Mode();
		final Mode m2 = new Mode();
		
		Vector<Mode> modes = new Vector<Mode>();
		modes.add(m1);
		modes.add(m2);
		//MTask(String name, long T, long Ca, long Cs, HashMap<Mode,Long> criticalities,long RMPriority, numModes)	
		MTask t1 = new MTask("t1",40, 1,1, new HashMap<Mode,Long> () {{put(m1,1l);put(m2,0l);}},3,2);
		MTask t2 = new MTask("t2",10, 5,1, new HashMap<Mode,Long> () {{put(m1,1l);}},1,2);
		MTask t3 = new MTask("t3",30, 3,1, new HashMap<Mode,Long> () {{put(m2,0l);}},2,2);
		
		
		
//		MTask t1 = new MTask("t1", 4, 2, 1,new HashMap<Mode,Long> (){{put(m1,1l);put(m2,0l);}},1,2);
//		MTask t2 = new MTask("t2", 8, 5, 2,new HashMap<Mode,Long> (){{put(m1,0l);put(m2,1l);}},2,2);
//		MTask t3 = new MTask("t3", 4, 2, 1,new HashMap<Mode,Long> (){{put(m1,1l);put(m2,0l);}},1,2);
//		MTask t4 = new MTask("t4", 8, 5, 2,new HashMap<Mode,Long> (){{put(m1,0l);put(m2,1l);}},2,2);

//		Vector<MTask> taskset = new Vector<MTask>();
//		taskset.add(t1);
//		taskset.add(t2);
//		taskset.add(t3);
//		taskset.add(t4);

		TransitionZSRMScheduler tscheduler = new TransitionZSRMScheduler();
		tscheduler.addTask(t1);
		tscheduler.addTask(t2);
		tscheduler.addTask(t3);
//		tscheduler.addTask(t1);
//		tscheduler.addTask(t2);
//		tscheduler.addTask(t3);
//		tscheduler.addTask(t4);
		
		long switchingPoint = tscheduler.getMaximizingMCRI(t1 , m1, m2);
		System.out.println(" max MCRI:"+switchingPoint);
		Vector<SlackRegion> srcRMSlackRegion  = tscheduler.calculateSlackVector(t1.getModalFacade(m1), new RMZonePlusHigherCriticalIterable(tscheduler.getModeTasksByPriority(m1),t1.getModalFacade(m1)));
		Vector<SlackRegion> tgtRMSlackRegion  = tscheduler.calculateSlackVector(t1.getModalFacade(m2), new RMZonePlusHigherCriticalIterable(tscheduler.getModeTasksByPriority(m2),t1.getModalFacade(m2)));
		Vector<SlackRegion> hybrid = tscheduler.buildHybridSlackVector(srcRMSlackRegion, tgtRMSlackRegion, switchingPoint);
		
		for (SlackRegion r:hybrid){
			System.out.println("time: "+r.getTimeInstant()+" slack: "+r.getSlack());
		}
	}
	
	// Test 
	public static void testCompression(){
		final Mode m1 = new Mode();
		final Mode m2 = new Mode();
		
		Vector<Mode> modes = new Vector<Mode>();
		modes.add(m1);
		modes.add(m2);
		//MTask(String name, long T, long Ca, long Cs, HashMap<Mode,Long> criticalities,long RMPriority)		
		MTask t1 = new MTask("t1", 4, 2, 1,new HashMap<Mode,Long> (){{put(m1,1l);put(m2,0l);}},1,2);
		MTask t2 = new MTask("t2", 8, 5, 2,new HashMap<Mode,Long> (){{put(m1,0l);put(m2,1l);}},2,2);
		MTask t3 = new MTask("t3", 4, 2, 1,new HashMap<Mode,Long> (){{put(m1,1l);put(m2,0l);}},1,2);
		MTask t4 = new MTask("t4", 8, 5, 2,new HashMap<Mode,Long> (){{put(m1,0l);put(m2,1l);}},2,2);

		Vector<MTask> taskset = new Vector<MTask>();
		taskset.add(t1);
		taskset.add(t2);
		taskset.add(t3);
		taskset.add(t4);

		TransitionZSRMScheduler tscheduler = new TransitionZSRMScheduler();
		tscheduler.addTask(t1);
		tscheduler.addTask(t2);
//		tscheduler.addTask(t3);
//		tscheduler.addTask(t4);
		
		tscheduler.TransitionCompressSchedule(m1,m2);
		
		System.out.println("t1.Z="+t1.transZeroSlackInstant);
		System.out.println("t2.Z="+t2.transZeroSlackInstant);
		double transduct = tscheduler.getTransitionalDuctility(m1,m2,2,new long[]{1,1},2);
		System.out.println("Tansitional ductility:"+transduct);
//		System.out.println("t3.Z="+t3.getModalFacade(m2).ZeroSlackInstant);
//		System.out.println("t4.Z="+t4.getModalFacade(m2).ZeroSlackInstant);
	}
	
	public static void main(String[] args){
		//testHybridSlackVector();
		testTransitionalCompress();
	}
}
